/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotSensorMemory
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// STD
#include <algorithm>

// Header
#include "RobotSensorMemory.h"

// Simox
#include <SimoxUtility/algorithm/string.h>

// ArmarX
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/server/MemoryRemoteGui.h>

#include <RobotAPI/libraries/aron/core/navigator/data/AllNavigators.h>
#include <RobotAPI/libraries/aron/core/navigator/type/AllNavigators.h>

namespace armarx
{
    RobotSensorMemory::RobotSensorMemory()
    {
    }

    armarx::PropertyDefinitionsPtr RobotSensorMemory::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs = new ComponentPropertyDefinitions(getConfigIdentifier());
        defs->optional(robotUnitSensorPrefix, "SensorValuePrefix", "Prefix of all sensor values");
        defs->optional(robotUnitMemoryBatchSize, "RobotUnitMemoryBatchSize", "The size of the entity snapshot to send to the memory. Min is 1");
        defs->optional(robotUnitPollFrequency, "RobotUnitUpdateFrequency", "The frequency in Hz to store values. All other values get discarded. Min is 1, max is " + std::to_string(ROBOT_UNIT_MAXIMUM_FREQUENCY));

        defs->optional(robotStateComponentMemoryBatchSize, "RobotStateComponentMemoryBatchSize", "The size of the entity snapshot to send to the memory. Min is 1");
        defs->optional(robotStateComponentPollFrequency, "RobotStateComponentUpdateFrequency", "The frequency in Hz to store values. All other values get discarded. Min is 1, max is " + std::to_string(ROBOT_STATE_COMPONENT_MAXIMUM_FREQUENCY));

        return defs;
    }


    std::string RobotSensorMemory::getDefaultName() const
    {
        return "RobotSensorMemory";
    }


    void RobotSensorMemory::onInitComponent()
    {
        robotUnitMemoryBatchSize = std::max((unsigned int) 1, robotUnitMemoryBatchSize);
        robotUnitPollFrequency = std::clamp(robotUnitPollFrequency, 1, ROBOT_UNIT_MAXIMUM_FREQUENCY);

        robotStateComponentMemoryBatchSize = std::max((unsigned int) 1, robotStateComponentMemoryBatchSize);
        robotStateComponentPollFrequency = std::clamp(robotStateComponentPollFrequency, 1, ROBOT_STATE_COMPONENT_MAXIMUM_FREQUENCY);

        workingMemory.name() = workingMemoryName;
    }


    void RobotSensorMemory::onConnectComponent()
    {
        setupRobotUnitSegment();
        setupRobotStateComponentSegment();

        cfg.loggingNames.emplace_back(robotUnitSensorPrefix);
        handler = make_shared<RobotUnitDataStreamingReceiver>(this, getRobotUnit(), cfg);
        keys = handler->getDataDescription();

        robotUnitConversionTask = new PeriodicTask<RobotSensorMemory>(this, &RobotSensorMemory::convertRobotUnitStreamingDataToAron, (1000 / robotUnitPollFrequency));
        robotUnitStoringTask = new PeriodicTask<RobotSensorMemory>(this, &RobotSensorMemory::storeConvertedRobotUnitDataInMemory, (1000 / robotUnitPollFrequency));

        robotStateComponentConversionTask = new PeriodicTask<RobotSensorMemory>(this, &RobotSensorMemory::convertRobotStateComponentDataToAron, (1000 / robotStateComponentPollFrequency));
        robotStateComponentStoringTask = new PeriodicTask<RobotSensorMemory>(this, &RobotSensorMemory::storeConvertedRobotStateComponentDataInMemory, (1000 / robotStateComponentPollFrequency));

        startRobotUnitStream();
        startRobotStateComponentPoll();
    }


    void RobotSensorMemory::onDisconnectComponent()
    {
        stopRobotUnitStream();
        startRobotStateComponentPoll();
    }


    void RobotSensorMemory::onExitComponent()
    {
        stopRobotUnitStream();
        stopRobotStateComponentPoll();
    }

    /*************************************************************/
    // RobotUnit Streaming functions
    /*************************************************************/
    void RobotSensorMemory::setupRobotUnitSegment()
    {
        ARMARX_INFO << "Adding core segment " << robotUnitCoreSegmentName;
        workingMemory.addCoreSegments({robotUnitCoreSegmentName});

        ARMARX_INFO << "Adding provider segment " << robotUnitCoreSegmentName << "/" << robotUnitProviderSegmentName;
        armem::data::AddSegmentInput input;
        input.coreSegmentName = robotUnitCoreSegmentName;
        input.providerSegmentName = robotUnitProviderSegmentName;

        auto encoderEntryType = std::make_shared<aron::typenavigator::ObjectNavigator>("RobotUnitEncoderEntry");
        auto encoderNameType = std::make_shared<aron::typenavigator::StringNavigator>();
        auto encoderIterationIDType = std::make_shared<aron::typenavigator::LongNavigator>();
        encoderEntryType->addMemberType("EncoderGroupName", encoderNameType);
        encoderEntryType->addMemberType("IterationId", encoderIterationIDType);
        //auto encoderValueType = std::make_shared<aron::typenavigator::AnyType>();
        //encoderEntryType->addMemberType("value", encoderValueType);

        auto result = addSegments({input})[0];

        if (!result.success)
        {
            ARMARX_ERROR << "Could not add segment " << robotUnitCoreSegmentName << "/" << robotUnitProviderSegmentName << ". The error message is: " << result.errorMessage;
        }

        robotUnitProviderID.memoryName = workingMemoryName;
        robotUnitProviderID.coreSegmentName = robotUnitCoreSegmentName;
        robotUnitProviderID.providerSegmentName = robotUnitProviderSegmentName;
    }

    void RobotSensorMemory::convertRobotUnitStreamingDataToAron()
    {
        auto& data = handler->getDataBuffer();
        if (data.size() == 0)
        {
            return;
        }
        auto& currentTimestep = data.back();

        ARMARX_DEBUG << "RobotUnitData: Generating new aron map for current timestep";
        auto start = std::chrono::high_resolution_clock::now();

        RobotUnitData convertedAndGroupedData;
        for (const auto& [name, dataEntry] : keys.entries)
        {
            size_t dot_pos = name.find(".", name.find(".") + 1); // find second occurence of "."
            if (dot_pos == std::string::npos)
            {
                ARMARX_WARNING << "Could not find a groupname for the sensor with name " << name << ". All sensors must be called sens.X.Y where X is the name of the group and Y is the actual sensor. Ignoring this sensor.";
                continue;
            }

            std::string groupName = name.substr(0, dot_pos);
            RobotUnitData::RobotUnitDataGroup e;
            if (auto it = convertedAndGroupedData.groups.find(groupName); it == convertedAndGroupedData.groups.end())
            {
                // generate new dict for the group
                auto dict = std::make_shared<aron::datanavigator::DictNavigator>();
                auto encGroupName = std::make_shared<aron::datanavigator::StringNavigator>(groupName);
                dict->addElement("EncoderGroupName", encGroupName);

                auto iterId = std::make_shared<aron::datanavigator::LongNavigator>(currentTimestep.iterationId);
                dict->addElement("IterationId", iterId);

                e.timestamp = currentTimestep.timestampUSec;
                e.name = groupName;
                e.data = dict;
            }
            else
            {
                // reuse existing entry
                e = it->second;
            }

            switch (dataEntry.type)
            {
                case RobotUnitDataStreaming::NodeTypeFloat:
                {
                    float value = RobotUnitDataStreamingReceiver::GetAs<Ice::Float>(currentTimestep, dataEntry);
                    auto aron = std::make_shared<aron::datanavigator::FloatNavigator>(value);
                    e.data->addElement(name, aron);
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeBool:
                {
                    bool value = RobotUnitDataStreamingReceiver::GetAs<bool>(currentTimestep, dataEntry);
                    auto aron = std::make_shared<aron::datanavigator::BoolNavigator>(value);
                    e.data->addElement(name, aron);
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeByte:
                {
                    int value = RobotUnitDataStreamingReceiver::GetAs<Ice::Byte>(currentTimestep, dataEntry);
                    auto aron = std::make_shared<aron::datanavigator::IntNavigator>(value);
                    e.data->addElement(name, aron);
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeShort:
                {
                    int value = RobotUnitDataStreamingReceiver::GetAs<Ice::Short>(currentTimestep, dataEntry);
                    auto aron = std::make_shared<aron::datanavigator::IntNavigator>(value);
                    e.data->addElement(name, aron);
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeInt:
                {
                    int value = RobotUnitDataStreamingReceiver::GetAs<Ice::Int>(currentTimestep, dataEntry);
                    auto aron = std::make_shared<aron::datanavigator::IntNavigator>(value);
                    e.data->addElement(name, aron);
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeLong:
                {
                    long value = RobotUnitDataStreamingReceiver::GetAs<Ice::Long>(currentTimestep, dataEntry);
                    auto aron = std::make_shared<aron::datanavigator::LongNavigator>(value);
                    e.data->addElement(name, aron);
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeDouble:
                {
                    double value = RobotUnitDataStreamingReceiver::GetAs<Ice::Double>(currentTimestep, dataEntry);
                    auto aron = std::make_shared<aron::datanavigator::DoubleNavigator>(value);
                    e.data->addElement(name, aron);
                    break;
                }
                default:
                    throw LocalException("The enum type should not exist! Perhaps someone changed the RobotUnitDataStreaming::NodeType enum definition?");
            }

            convertedAndGroupedData.groups.insert({groupName, e});
        }
        {
            std::lock_guard g{robotUnitDataMutex};
            robotUnitDataQueue.push(convertedAndGroupedData);
        }

        auto stop = std::chrono::high_resolution_clock::now();
        ARMARX_DEBUG << "RobotUnitData: The total time needed to convert the data to aron is: " << std::chrono::duration_cast<std::chrono::microseconds>(stop - start);
    }

    void RobotSensorMemory::storeConvertedRobotUnitDataInMemory()
    {
        unsigned int size = 0;
        {
            std::lock_guard g{robotUnitDataMutex};
            size = robotUnitDataQueue.size(); // the size can only grow
        }
        if (size >= robotUnitMemoryBatchSize)
        {
            ARMARX_DEBUG << "RobotUnitData: Sending batch of " << robotUnitMemoryBatchSize << " timesteps to memory... The size of the queue is: " << size;
            auto start = std::chrono::high_resolution_clock::now();

            // send batch to memory
            armem::data::Commit c;
            for (unsigned int i = 0; i < robotUnitMemoryBatchSize; ++i)
            {
                std::lock_guard g{robotUnitDataMutex};
                const auto& convertedAndGroupedData = robotUnitDataQueue.front();
                for (const auto& [encName, encTimestep] : convertedAndGroupedData.groups)
                {
                    ARMARX_CHECK_EQUAL(encName, encTimestep.name);

                    auto entityID = robotUnitProviderID;
                    entityID.entityName = encName;

                    const auto& timeUSec = encTimestep.timestamp;
                    const auto& aron = encTimestep.data;

                    armem::data::EntityUpdate& update = c.updates.emplace_back();
                    update.entityID = entityID;
                    update.timeCreatedMicroSeconds = timeUSec;
                    update.instancesData = {aron->toAronDictPtr()};
                }
                robotUnitDataQueue.pop();
            }

            auto results = commit(c);
            auto stop = std::chrono::high_resolution_clock::now();
            ARMARX_DEBUG << "RobotUnitData: The total runtime of sending a batch to the memory is: " << std::chrono::duration_cast<std::chrono::microseconds>(stop - start);

            for (const auto& result : results.results)
            {
                if (!result.success)
                {
                    ARMARX_WARNING << "Could not add data to memory. Error message: " << result.errorMessage;
                }
            }
        }
    }

    void RobotSensorMemory::stopRobotUnitStream()
    {
        std::lock_guard g{startStopMutex};
        robotUnitConversionTask->stop();
        robotUnitStoringTask->stop();
    }

    void RobotSensorMemory::startRobotUnitStream()
    {
        std::lock_guard g{startStopMutex};
        if (robotUnitConversionTask->isRunning() || robotUnitStoringTask->isRunning())
        {
            if (robotUnitConversionTask->isRunning() && robotUnitStoringTask->isRunning())
            {
                return;
            }
            ARMARX_WARNING << "Found inconsistency in running tasks. Restarting all!";
            stopRobotUnitStream();
        }

        std::stringstream ss;
        ss << "Getting sensor values for:" << std::endl;
        for (const auto& [name, dataEntry] : keys.entries)
        {
            std::string type = "";
            switch (dataEntry.type)
            {
                case RobotUnitDataStreaming::NodeTypeBool:
                {
                    type = "Bool";
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeByte:
                {
                    type = "Byte";
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeShort:
                {
                    type = "Short";
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeInt:
                {
                    type = "Int";
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeLong:
                {
                    type = "Long";
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeFloat:
                {
                    type = "Float";
                    break;
                }
                case RobotUnitDataStreaming::NodeTypeDouble:
                {
                    type = "Double";
                    break;
                }
            }
            ss << "\t" << name << " (type: '" << type << "')" << std::endl;
        }
        ARMARX_INFO << ss.str();

        robotUnitConversionTask->start();
        robotUnitStoringTask->start();
    }


    /*************************************************************/
    // RobotStateComponent Polling functions
    /*************************************************************/
    void RobotSensorMemory::setupRobotStateComponentSegment()
    {
        ARMARX_INFO << "Adding core segment " << robotStateComponentCoreSegmentName;
        workingMemory.addCoreSegments({robotStateComponentCoreSegmentName});

        ARMARX_INFO << "Adding provider segment " << robotStateComponentCoreSegmentName << "/" << robotStateComponentProviderSegmentName;
        armem::data::AddSegmentInput input;
        input.coreSegmentName = robotStateComponentCoreSegmentName;
        input.providerSegmentName = robotStateComponentProviderSegmentName;

        // Todo add type

        auto result = addSegments({input})[0];

        if (!result.success)
        {
            ARMARX_ERROR << "Could not add segment " << robotStateComponentCoreSegmentName << "/" << robotStateComponentProviderSegmentName << ". The error message is: " << result.errorMessage;
        }

        robotStateComponentProviderID.memoryName = workingMemoryName;
        robotStateComponentProviderID.coreSegmentName = robotStateComponentCoreSegmentName;
        robotStateComponentProviderID.providerSegmentName = robotStateComponentProviderSegmentName;
    }

    void RobotSensorMemory::convertRobotStateComponentDataToAron()
    {
        ARMARX_DEBUG << "RobotStateComponentData: Generating new aron map for current timestep";
        auto start = std::chrono::high_resolution_clock::now();

        auto synchronizedRobot = getRobotStateComponent()->getSynchronizedRobot();
        auto framedPose = synchronizedRobot->getRootNode()->getGlobalPose();

        RobotStateComponentData convertedData;
        convertedData.timestamp = synchronizedRobot->getTimestamp()->timestamp;

        auto dict = std::make_shared<aron::datanavigator::DictNavigator>();

        auto x = std::make_shared<aron::datanavigator::IntNavigator>(framedPose->position->x);
        dict->addElement("x", x);

        auto y = std::make_shared<aron::datanavigator::IntNavigator>(framedPose->position->y);
        dict->addElement("y", y);

        auto z = std::make_shared<aron::datanavigator::IntNavigator>(framedPose->position->z);
        dict->addElement("z", z);

        auto qw = std::make_shared<aron::datanavigator::IntNavigator>(framedPose->orientation->qw);
        dict->addElement("qw", qw);

        auto qx = std::make_shared<aron::datanavigator::IntNavigator>(framedPose->orientation->qx);
        dict->addElement("qx", qx);

        auto qy = std::make_shared<aron::datanavigator::IntNavigator>(framedPose->orientation->qy);
        dict->addElement("qy", qy);

        auto qz = std::make_shared<aron::datanavigator::IntNavigator>(framedPose->orientation->qz);
        dict->addElement("z", qz);

        convertedData.data = dict;

        {
            std::lock_guard g{robotStateComponentDataMutex};
            robotStateComponentDataQueue.push(convertedData);
        }

        auto stop = std::chrono::high_resolution_clock::now();
        ARMARX_DEBUG << "RobotStateComponentData: The total time needed to convert the data to aron is: " << std::chrono::duration_cast<std::chrono::microseconds>(stop - start);
    }

    void RobotSensorMemory::storeConvertedRobotStateComponentDataInMemory()
    {
        unsigned int size = 0;
        {
            std::lock_guard g{robotStateComponentDataMutex};
            size = robotStateComponentDataQueue.size(); // the size can only grow
        }
        if (size >= robotStateComponentMemoryBatchSize)
        {
            ARMARX_DEBUG << "RobotStateComponentData: Sending batch of " << robotStateComponentMemoryBatchSize << " timesteps to memory... The size of the queue is: " << size;
            auto start = std::chrono::high_resolution_clock::now();

            // send batch to memory
            armem::data::Commit c;
            for (unsigned int i = 0; i < robotStateComponentMemoryBatchSize; ++i)
            {
                std::lock_guard g{robotStateComponentDataMutex};
                const auto& convertedData = robotStateComponentDataQueue.front();

                auto entityID = robotStateComponentProviderID;
                entityID.entityName = "GlobalPositionRootFrame";

                const auto& timeUSec = convertedData.timestamp;
                const auto& aron = convertedData.data;

                armem::data::EntityUpdate& update = c.updates.emplace_back();
                update.entityID = entityID;
                update.timeCreatedMicroSeconds = timeUSec;
                update.instancesData = {aron->toAronDictPtr()};
                robotStateComponentDataQueue.pop();
            }

            auto results = commit(c);
            auto stop = std::chrono::high_resolution_clock::now();
            ARMARX_DEBUG << "RobotStateComponentData: The total runtime of sending a batch to the memory is: " << std::chrono::duration_cast<std::chrono::microseconds>(stop - start);

            for (const auto& result : results.results)
            {
                if (!result.success)
                {
                    ARMARX_WARNING << "Could not add data to memory. Error message: " << result.errorMessage;
                }
            }
        }
    }

    void RobotSensorMemory::stopRobotStateComponentPoll()
    {
        std::lock_guard g{startStopMutex};
        robotStateComponentConversionTask->stop();
        robotStateComponentStoringTask->stop();
    }

    void RobotSensorMemory::startRobotStateComponentPoll()
    {
        std::lock_guard g{startStopMutex};
        if (robotStateComponentConversionTask->isRunning() || robotStateComponentStoringTask->isRunning())
        {
            if (robotStateComponentConversionTask->isRunning() && robotStateComponentStoringTask->isRunning())
            {
                return;
            }
            ARMARX_WARNING << "Found inconsistency in running tasks. Restarting all!";
            stopRobotStateComponentPoll();
        }

        ARMARX_INFO << "Start polling of robotStateComponent";

        robotStateComponentConversionTask->start();
        robotStateComponentStoringTask->start();
    }
}
