#include "MemoryToIceAdapter.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "../error.h"
#include "../core/workingmemory/ice_conversions.h"
#include "query_proc/workingmemory/MemoryQueryProcessor.h"

//#include "../core/io/diskWriter/NlohmannJSON/NlohmannJSONDiskWriter.h"
//#include "../core/io/diskReader/NlohmannJSON/NlohmannJSONDiskReader.h"

namespace armarx::armem::server
{

    MemoryToIceAdapter::MemoryToIceAdapter(wm::Memory* workingMemory, ltm::Memory* longtermMemory) :
        workingMemory(workingMemory), longtermMemory(longtermMemory)
    {
    }


    void
    MemoryToIceAdapter::setMemoryListener(client::MemoryListenerInterfacePrx memoryListener)
    {
        this->memoryListenerTopic = memoryListener;
    }


    // WRITING
    data::AddSegmentResult
    MemoryToIceAdapter::addSegment(const data::AddSegmentInput& input, bool addCoreSegments)
    {
        ARMARX_DEBUG << "Adding segment '" << input.coreSegmentName << "/" << input.providerSegmentName << "'.";
        ARMARX_CHECK_NOT_NULL(workingMemory);

        data::AddSegmentResult output;

        armem::wm::CoreSegment* coreSegment = nullptr;
        try
        {
            coreSegment = &workingMemory->getCoreSegment(input.coreSegmentName);
        }
        catch (const armem::error::MissingEntry& e)
        {
            if (addCoreSegments)
            {
                coreSegment = &workingMemory->addCoreSegment(input.coreSegmentName);
            }
            else
            {
                output.success = false;
                output.errorMessage = e.what();
                return output;
            }
        }
        ARMARX_CHECK_NOT_NULL(coreSegment);

        if (input.providerSegmentName.size() > 0)
        {
            try
            {
                coreSegment->addProviderSegment(input.providerSegmentName);
            }
            catch (const armem::error::ContainerEntryAlreadyExists&)
            {
                // This is ok.
                if (input.clearWhenExists)
                {
                    wm::ProviderSegment& provider = coreSegment->getProviderSegment(input.providerSegmentName);
                    provider.clear();
                }
            }
        }

        armem::MemoryID segmentID;
        segmentID.memoryName = workingMemory->name();
        segmentID.coreSegmentName = input.coreSegmentName;
        segmentID.providerSegmentName = input.providerSegmentName;

        output.success = true;
        output.segmentID = segmentID.str();
        return output;
    }


    data::AddSegmentsResult
    MemoryToIceAdapter::addSegments(const data::AddSegmentsInput& input, bool addCoreSegments)
    {
        ARMARX_CHECK_NOT_NULL(workingMemory);

        data::AddSegmentsResult output;
        for (const auto& i : input)
        {
            output.push_back(addSegment(i, addCoreSegments));
        }
        return output;
    }

    data::CommitResult
    MemoryToIceAdapter::commit(const data::Commit& commitIce, Time timeArrived)
    {
        ARMARX_CHECK_NOT_NULL(workingMemory);

        armem::Commit commit;
        armem::fromIce(commitIce, commit, timeArrived);

        armem::CommitResult result = this->commit(commit);
        data::CommitResult resultIce;
        armem::toIce(resultIce, result);

        return resultIce;
    }


    data::CommitResult
    MemoryToIceAdapter::commit(const data::Commit& commitIce)
    {
        return commit(commitIce, armem::Time::now());
    }


    armem::CommitResult
    MemoryToIceAdapter::commit(const armem::Commit& commit)
    {
        std::vector<data::MemoryID> updatedIDs;
        const bool publishUpdates = memoryListenerTopic;

        CommitResult commitResult;
        for (const EntityUpdate& update : commit.updates)
        {
            EntityUpdateResult& result = commitResult.results.emplace_back();
            try
            {
                MemoryID snapshotID = workingMemory->update(update);

                result.success = true;
                result.snapshotID = snapshotID;
                result.timeArrived = update.timeArrived;

                if (publishUpdates)
                {
                    data::MemoryID& id = updatedIDs.emplace_back();
                    armem::toIce(id, snapshotID);
                }
            }
            catch (const error::ArMemError& e)
            {
                result.success = false;
                result.errorMessage = e.what();
            }
        }

        if (publishUpdates)
        {
            memoryListenerTopic->memoryUpdated(updatedIDs);
        }

        return commitResult;
    }


    // READING
    armem::query::data::Result
    MemoryToIceAdapter::query(const armem::query::data::Input& input)
    {
        ARMARX_CHECK_NOT_NULL(workingMemory);

        armem::wm::query_proc::MemoryQueryProcessor processor(
            input.withData ? armem::DataMode::WithData
            : armem::DataMode::NoData);
        return processor.process(input, *workingMemory);
    }

    // LTM LOADING
    query::data::Result MemoryToIceAdapter::load(const armem::query::data::Input& query)
    {
        ARMARX_CHECK_NOT_NULL(longtermMemory);
        query::data::Result output;

        output.success = true;
        return output;
    }

    // LTM STORING
    data::StoreResult MemoryToIceAdapter::store(const armem::data::StoreInput& input)
    {
        ARMARX_CHECK_NOT_NULL(workingMemory);
        ARMARX_CHECK_NOT_NULL(longtermMemory);
        data::StoreResult output;

        armem::query::data::Result queryResult = this->query(input.query);
        if (queryResult.success)
        {
            wm::Memory m;
            fromIce(queryResult.memory, m);
            longtermMemory->append(m);
        }
        else
        {
            output.success = false;
        }

        ARMARX_IMPORTANT << "Finsihed";
        return output;
    }

    client::QueryResult MemoryToIceAdapter::query(const client::QueryInput& input)
    {
        return client::QueryResult::fromIce(query(input.toIce()));
    }

}
