#include "Segment.h"

#include <RobotAPI/libraries/aron/core/Exception.h>
#include <RobotAPI/libraries/aron/common/aron_conversions.h>
#include <RobotAPI/libraries/ArmarXObjects/aron_conversions.h>
#include <RobotAPI/libraries/armem_objects/aron_conversions.h>

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/time/TimeUtil.h>

#include <SimoxUtility/color/Color.h>
#include <SimoxUtility/math/pose/pose.h>
#include <SimoxUtility/shapes/AxisAlignedBoundingBox.h>
#include <SimoxUtility/shapes/OrientedBox.h>

#include <filesystem>


namespace armarx::armem::server::obj::clazz
{

    Segment::Segment(armem::server::MemoryToIceAdapter& memoryToIceAdapter, std::mutex& memoryMutex) :
        iceMemory(memoryToIceAdapter),
        memoryMutex(memoryMutex)
    {
        Logging::setTag("ClassSegment");
    }

    void Segment::defineProperties(armarx::PropertyDefinitionsPtr defs, const std::string& prefix)
    {
        defs->optional(p.coreSegmentName, prefix + "CoreSegmentName", "Name of the object clazz core segment.");
        defs->optional(p.maxHistorySize, prefix + "MaxHistorySize", "Maximal size of object poses history (-1 for infinite).");

        defs->optional(p.objectsPackage, prefix + "ObjectsPackgage", "Name of the objects package to load from.");
        defs->optional(p.loadFromObjectsPackage, prefix + "LoadFromObjectsPackage",
                       "If true, load the objects from the objects package on startup.");
    }

    void Segment::init()
    {
        ARMARX_CHECK_NOT_NULL(iceMemory.workingMemory);

        coreSegment = &iceMemory.workingMemory->addCoreSegment(p.coreSegmentName, arondto::ObjectClass::toInitialAronType());
        coreSegment->setMaxHistorySize(p.maxHistorySize);

        if (p.loadFromObjectsPackage)
        {
            loadByObjectFinder(p.objectsPackage);
        }
    }

    void Segment::connect(viz::Client arviz)
    {
        this->arviz = arviz;
    }


    void Segment::loadByObjectFinder(const std::string& objectsPackage)
    {
        loadByObjectFinder(ObjectFinder(objectsPackage));
    }

    void Segment::loadByObjectFinder(const ObjectFinder& finder)
    {
        this->objectFinder = finder;
        loadByObjectFinder();
    }

    void Segment::loadByObjectFinder()
    {
        const Time now = TimeUtil::GetTime();

        const bool checkPaths = false;
        std::vector<ObjectInfo> infos = objectFinder.findAllObjects(checkPaths);

        const MemoryID providerID = coreSegment->id().withProviderSegmentName(objectFinder.getPackageName());
        coreSegment->addProviderSegment(providerID.providerSegmentName);

        ARMARX_INFO << "Loading up to " << infos.size() << " object classes from '"
                    << objectFinder.getPackageName() << "' ...";
        Commit commit;
        for (ObjectInfo& info : infos)
        {
            info.setLogError(false);

            EntityUpdate& update = commit.updates.emplace_back();
            update.entityID = providerID.withEntityName(info.id().str());
            update.timeArrived = update.timeCreated = update.timeSent = now;

            arondto::ObjectClass objectClass = objectClassFromInfo(info);
            update.instancesData =
            {
                objectClass.toAron()
            };
        }
        ARMARX_INFO << "Loaded " << commit.updates.size() << " object classes from '"
                    << objectFinder.getPackageName() << "'.";
        iceMemory.commit(commit);
    }

    void Segment::visualizeClass(const MemoryID& entityID, bool showAABB, bool showOOBB)
    {
        const Eigen::Matrix4f pose = Eigen::Matrix4f::Identity();

        viz::Layer layerOrigin = arviz.layer("Origin");
        layerOrigin.add(viz::Pose("Origin"));

        viz::Layer layerObject = arviz.layer("Class Model");
        viz::Layer layerAABB = arviz.layer("Class AABB");
        viz::Layer layerOOBB = arviz.layer("Class OOBB");

        if (coreSegment)
        {
            try
            {
                const armem::wm::Entity& entity = coreSegment->getEntity(entityID);
                const armem::wm::EntityInstance& instance = entity.getLatestSnapshot().getInstance(0);

                arondto::ObjectClass aron;
                aron.fromAron(instance.data());

                if (!aron.simoxXmlPath.package.empty())
                {
                    layerObject.add(viz::Object(entityID.str())
                                    .file(aron.simoxXmlPath.package, aron.simoxXmlPath.path)
                                    .pose(pose));
                }

                if (showAABB)
                {
                    layerAABB.add(viz::Box("AABB")
                                  .pose(pose * simox::math::pose(aron.aabb.center))
                                  .size(aron.aabb.extents)
                                  .color(simox::Color::cyan(255, 64)));
                }
                if (showOOBB)
                {
                    layerOOBB.add(viz::Box("OOBB")
                                  .pose(pose * simox::math::pose(aron.oobb.center, aron.oobb.orientation))
                                  .size(aron.oobb.extents)
                                  .color(simox::Color::lime(255, 64)));
                }
            }
            catch (const armem::error::ArMemError& e)
            {
                ARMARX_INFO << "Failed to visualize object class " << entityID << "."
                            << "\nReason: " << e.what();
            }
            catch (const aron::error::AronException& e)
            {
                ARMARX_INFO << "Failed to visualize object class " << entityID << "."
                            << "\nReason: " << e.what();
            }
        }

        arviz.commit(layerObject, layerOrigin, layerAABB, layerOOBB);
    }

    arondto::ObjectClass Segment::objectClassFromInfo(const ObjectInfo& info)
    {
        namespace fs = std::filesystem;

        arondto::ObjectClass data;

        toAron(data.id, info.id());

        if (fs::is_regular_file(info.simoxXML().absolutePath))
        {
            toAron(data.simoxXmlPath, info.simoxXML());
        }
        if (fs::is_regular_file(info.wavefrontObj().absolutePath))
        {
            toAron(data.meshObjPath, info.wavefrontObj());
        }
        if (fs::is_regular_file(info.file(".wrl").absolutePath))
        {
            toAron(data.meshWrlPath, info.file(".wrl"));
        }

        auto aabb = info.loadAABB();
        toAron(data.aabb, aabb ? aabb.value() : simox::AxisAlignedBoundingBox());
        auto oobb = info.loadOOBB();
        toAron(data.oobb, oobb ? oobb.value() : simox::OrientedBoxf());

        if (auto recogNames = info.loadRecognizedNames())
        {
            data.names.recognizedNames = recogNames.value();
        }
        if (auto spokenNames = info.loadSpokenNames())
        {
            data.names.spokenNames = spokenNames.value();
        }

        return data;
    }


    void Segment::RemoteGui::setup(const Segment& segment)
    {
        using namespace armarx::RemoteGui::Client;

        data.setup(segment);
        visu.setup(segment);

        VBoxLayout layout;
        layout.addChildren({data.group, visu.group});

        group = {};
        group.setLabel("Class");
        group.addChildren({layout, VSpacer()});
    }

    void Segment::RemoteGui::update(Segment& segment)
    {
        data.update(segment);
        visu.update(segment);
    }

    void Segment::RemoteGui::Data::setup(const Segment& segment)
    {
        using namespace armarx::RemoteGui::Client;

        maxHistorySize.setValue(std::max(1, int(segment.p.maxHistorySize)));
        maxHistorySize.setRange(1, 1e6);
        infiniteHistory.setValue(segment.p.maxHistorySize == -1);

        GridLayout grid;
        int row = 0;
        grid.add(Label("Max History Size"), {row, 0}).add(maxHistorySize, {row, 1});
        row++;
        grid.add(Label("Infinite History Size"), {row, 0}).add(infiniteHistory, {row, 1});
        row++;

        group = {};
        group.setLabel("Data");
        group.addChild(grid);
    }

    void Segment::RemoteGui::Data::update(Segment& segment)
    {
        if (infiniteHistory.hasValueChanged() || maxHistorySize.hasValueChanged())
        {
            std::scoped_lock lock(segment.memoryMutex);
            segment.p.maxHistorySize = infiniteHistory.getValue() ? -1 : maxHistorySize.getValue();
            if (segment.coreSegment)
            {
                segment.coreSegment->setMaxHistorySize(long(segment.p.maxHistorySize));
            }
        }
    }



    void Segment::RemoteGui::Visu::setup(const Segment& segment)
    {
        using namespace armarx::RemoteGui::Client;

        showComboBox = {};
        showOptionsIndex.clear();
        for (const auto& [_, prov] : *segment.coreSegment)
        {
            for (const auto& [_, entity] : prov)
            {
                std::stringstream option;
                option << entity.id().entityName << " (" << entity.id().providerSegmentName << ")";
                showComboBox.addOption(option.str());
                showOptionsIndex.push_back(entity.id());
            }
        }
        if (showOptionsIndex.empty())
        {
            showComboBox.addOption("<none>");
        }
        showButton.setLabel("Visualize Object Class");

        GridLayout grid;
        int row = 0;
        grid.add(showComboBox, {row, 0}, {1, 2});
        row++;
        grid.add(showButton, {row, 0}, {1, 2});
        row++;

        group = {};
        group.setLabel("Visualization");
        group.addChild(grid);
    }

    void Segment::RemoteGui::Visu::update(Segment& segment)
    {
        if (showButton.wasClicked())
        {
            const size_t index = static_cast<size_t>(showComboBox.getIndex());
            if (/*index >= 0 &&*/ index < showOptionsIndex.size())
            {
                std::scoped_lock lock(segment.memoryMutex);
                segment.visualizeClass(showOptionsIndex.at(index));
            }
        }
    }

}
