/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "TransformReader.h"

#include <Eigen/src/Geometry/Transform.h>
#include <algorithm>
#include <iterator>
#include <numeric>

// Ice
#include <IceUtil/Time.h>

// Eigen
#include <Eigen/Geometry>

// Simox
#include <SimoxUtility/algorithm/get_map_keys_values.h>
#include <SimoxUtility/algorithm/string/string_tools.h>
#include <SimoxUtility/color/cmaps.h>
#include <SimoxUtility/math/pose/interpolate.h>

// ArmarX
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/time/CycleUtil.h>

// this package
#include <RobotAPI/libraries/armem/client/query/Builder.h>
#include <RobotAPI/libraries/armem/client/query/query_fns.h>
#include <RobotAPI/libraries/armem/core/workingmemory/Memory.h>
#include <RobotAPI/libraries/armem/core/workingmemory/ProviderSegment.h>
#include <RobotAPI/libraries/armem/core/workingmemory/ice_conversions.h>
#include <RobotAPI/libraries/armem/server/MemoryRemoteGui.h>
#include <RobotAPI/libraries/aron/core/navigator/type/NavigatorFactory.h>
#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/libraries/armem_robot_localization/aron_conversions.h>
#include <RobotAPI/libraries/armem_robot_localization/aron/Transform.aron.generated.h>

namespace armarx::armem
{

    TransformReader::TransformReader(ManagedIceObject& component) : MemoryConnector(component) {}

    TransformReader::~TransformReader() = default;

    void TransformReader::registerPropertyDefinitions(armarx::PropertyDefinitionsPtr& def)
    {
        ARMARX_DEBUG << "TransformReader: registerPropertyDefinitions";
        MemoryConnector::registerPropertyDefinitions(def);

        const std::string prefix = getPropertyPrefix();

        def->optional(properties.localizationMemoryName,
                      prefix + "LocalizationMemoryName",
                      "Name of the localization memory core segment to use.");

        def->optional(properties.memoryName, prefix + "MemoryName");
    }

    void TransformReader::connect()
    {
        // Wait for the memory to become available and add it as dependency.
        ARMARX_IMPORTANT << "TransformReader: Waiting for memory '" << properties.memoryName
                         << "' ...";
        auto result = useMemory(properties.memoryName);
        if (not result.success)
        {
            ARMARX_ERROR << result.errorMessage;
            return;
        }

        ARMARX_IMPORTANT << "TransformReader: Connected to memory '" << properties.memoryName;

        memoryReader.setReadingMemory(result.proxy);
    }

    TransformResult TransformReader::getGlobalPose(const std::string& agentName,
            const std::string& robotRootFrame,
            const std::int64_t& timestamp) const
    {
        const TransformQuery query{.header = {.parentFrame = GlobalFrame,
                                              .frame       = robotRootFrame,
                                              .agent       = agentName,
                                              .timestamp   = timestamp
                                             }};

        return lookupTransform(query);
    }

    // std::vector<std::string> buildTransformChainArbitrary(const
    // std::map<std::string, armem::ProviderSegment>& providerSegments, const
    // std::string& parentFrame, const std::string& frame){

    //     std::map<std::string, std::string> tf;

    //     // build string representation of transforms
    //     for(const auto& [name, segment]: providerSegments){
    //         const auto frames = simox::alg::split(name, ",");
    //         if (frames.size() != 2){
    //             ARMARX_WARNING << "Segment name " << name << " not understood";
    //             continue;
    //         }

    //         tf.insert(frame[0], frame[1]);
    //     }

    //     // find

    // }

    std::vector<std::string> TransformReader::buildTransformChain(const armem::wm::Memory& memory,
            const TransformQuery& query) const
    {
        ARMARX_DEBUG << "Building transform chain";

        auto join = [](const std::string & parentFrame, const std::string & frame)
        {
            return parentFrame + "," + frame;
        };

        std::vector<std::string> chain;

        const auto& agentProviderSegment = memory.getCoreSegment(properties.localizationMemoryName)
                                           .getProviderSegment(query.header.agent);

        auto addToChain = [&](const std::string & parentFrame, const std::string & frame)
        {
            const auto entityName = join(parentFrame, frame);

            if (agentProviderSegment.hasEntity(entityName))
            {
                chain.push_back(entityName);
            }
            else
            {
                ARMARX_WARNING << "Cannot perform tf lookup '" << parentFrame << " -> " << frame
                               << "'";
            }
        };

        std::array<std::string, 3> knownChain
        {
            GlobalFrame, MapFrame, OdometryFrame}; // Robot comes next

        auto* frameBeginIt =
            std::find(knownChain.begin(), knownChain.end(), query.header.parentFrame);
        auto* const frameEndIt =
            std::find(knownChain.begin(), knownChain.end(), query.header.frame);

        if (frameBeginIt == knownChain.end())
        {
            ARMARX_WARNING << "Parent frame '" << query.header.parentFrame << "' unknown";
            return {};
        }

        if (frameEndIt == knownChain.end())
        {
            ARMARX_DEBUG << "Frame '" << query.header.frame << "' must be robot frame";
        }

        const size_t nFrames = std::distance(frameBeginIt, frameEndIt);
        ARMARX_DEBUG << "Lookup '" << query.header.parentFrame << " -> " << query.header.frame
                     << "' across " << nFrames << " frames";

        for (; frameBeginIt != knownChain.end() - 1; frameBeginIt++)
        {
            addToChain(*frameBeginIt, *(frameBeginIt + 1));
        }

        if (frameEndIt == knownChain.end())
        {
            addToChain(knownChain.back(), query.header.frame);
        }

        if (chain.empty())
        {
            ARMARX_WARNING << "Cannot create tf lookup chain '" << query.header.parentFrame
                           << " -> " << query.header.frame << "'";
            return {};
        }

        return chain;
    }

    inline Transform convertEntityToTransform(const armem::wm::EntityInstance& item)
    {
        aron::Transform aronTransform;
        aronTransform.fromAron(item.data());

        Transform transform;
        fromAron(aronTransform, transform);

        return transform;
    }

    decltype(auto) findFirstElementAfter(const std::vector<Transform>& transforms,
                                         const int64_t timestamp)
    {
        auto timestampBeyond = [timestamp](const Transform & transform)
        {
            return transform.header.timestamp > timestamp;
        };

        const auto poseNextIt = std::find_if(transforms.begin(), transforms.end(), timestampBeyond);
        return poseNextIt;
    }

    Eigen::Affine3f interpolateTransform(const std::vector<Transform>& queue, int64_t timestamp)
    {
        ARMARX_TRACE;

        ARMARX_DEBUG << "Entering";


        ARMARX_CHECK(not queue.empty())
                << "The queue has to contain at least two items to perform a lookup";


        ARMARX_DEBUG << "Entering ... "
                     << "Q front " << queue.front().header.timestamp << "  "
                     << "Q back " << queue.back().header.timestamp << "  "
                     << "query timestamp " << timestamp;

        // TODO(fabian.reister): sort queue.


        ARMARX_CHECK(queue.back().header.timestamp > timestamp)
                << "Cannot perform lookup into the future!";

        // ARMARX_DEBUG << "Entering 1.5 " << queue.front().timestamp << "  " << timestamp;
        ARMARX_CHECK(queue.front().header.timestamp < timestamp)
                << "Cannot perform lookup. Timestamp too old";
        // => now we know that there is an element right after and before the timestamp within our queue

        ARMARX_DEBUG << "Entering 2";

        const auto poseNextIt = findFirstElementAfter(queue, timestamp);

        ARMARX_DEBUG << "it ari";

        const auto posePreIt = poseNextIt - 1;

        ARMARX_DEBUG << "deref";

        // the time fraction [0..1] of the lookup wrt to posePre and poseNext
        const float t = static_cast<float>(timestamp - posePreIt->header.timestamp) /
                        (poseNextIt->header.timestamp - posePreIt->header.timestamp);

        ARMARX_DEBUG << "interpolate";

        return simox::math::interpolatePose(posePreIt->transform, poseNextIt->transform, t);
    }

    Eigen::Affine3f TransformReader::obtainTransform(const std::string& entityName, const armem::wm::ProviderSegment& agentProviderSegment, const int64_t timestamp) const
    {

        ARMARX_DEBUG << "getEntity:" + entityName;
        const auto& entity = agentProviderSegment.getEntity(entityName);

        ARMARX_DEBUG << "History (size: " << entity.history().size() << ")"
                     << simox::alg::get_keys(entity.history());

        // if (entity.history.empty())
        // {
        //     // TODO(fabian.reister): fixme boom
        //     ARMARX_ERROR << "No snapshots received.";
        //     return Eigen::Affine3f::Identity();
        // }


        std::vector<Transform> transforms;
        transforms.reserve(entity.history().size());

        const auto entitySnapshots = simox::alg::get_values(entity.history());
        std::transform(entitySnapshots.begin(),
                       entitySnapshots.end(),
                       std::back_inserter(transforms),
                       [](const auto & entity)
        {
            return convertEntityToTransform(entity.getInstance(0));
        });

        ARMARX_DEBUG << "obtaining transform";

        if (transforms.size() > 1)
        {
            // TODO(fabian.reister): remove
            return transforms.front().transform;

            ARMARX_DEBUG << "More than one snapshots received: " << transforms.size();
            const auto p = interpolateTransform(transforms, timestamp);
            ARMARX_DEBUG << "Done interpolating transform";
            return p;
        }


        // accept this to fail (will raise armem::error::MissingEntry)
        if (transforms.empty())
        {
            ARMARX_DEBUG << "empty transform";

            throw armem::error::MissingEntry("foo", "bar", "foo2", "bar2");
        }

        ARMARX_DEBUG << "single transform";


        return transforms.front().transform;
    }

    std::vector<Eigen::Affine3f>
    TransformReader::obtainTransforms(const armem::wm::Memory& memory,
                                      const std::vector<std::string>& tfChain,
                                      const std::string& agent,
                                      const std::int64_t& timestamp) const
    {

        ARMARX_DEBUG << "Core segments" << simox::alg::get_keys(memory.coreSegments());

        const auto& agentProviderSegment =
            memory.getCoreSegment(properties.localizationMemoryName).getProviderSegment(agent);

        ARMARX_DEBUG << "Provider segments"
                     << simox::alg::get_keys(
                         memory.getCoreSegment(properties.localizationMemoryName)
                         .providerSegments());

        ARMARX_DEBUG << "Entities: " << simox::alg::get_keys(agentProviderSegment.entities());

        try
        {
            std::vector<Eigen::Affine3f> transforms;
            transforms.reserve(tfChain.size());
            std::transform(tfChain.begin(),
                           tfChain.end(),
                           std::back_inserter(transforms),
                           [&](const std::string & entityName)
            {
                return obtainTransform(entityName, agentProviderSegment, timestamp);
            });
            return transforms;
        }
        catch (const armem::error::MissingEntry& missingEntryError)
        {
            ARMARX_WARNING << missingEntryError.what();
        }
        catch (const ::armarx::exceptions::local::ExpressionException& ex)
        {
            ARMARX_WARNING << "local exception: " <<  ex.what();
        }

        return {};

    }

    TransformResult TransformReader::lookupTransform(const TransformQuery& query) const
    {

        const auto timestamp = IceUtil::Time::microSeconds(query.header.timestamp);

        const auto durationEpsilon = IceUtil::Time::milliSeconds(100);

        ARMARX_DEBUG << "Looking up transform at timestamp " << timestamp;

        // Query all entities from provider.
        armem::client::query::Builder qb;

        // clang-format off
        qb
        .coreSegments().withName(properties.localizationMemoryName)
        .providerSegments().withName(query.header.agent) // agent
        .entities().all() // parentFrame,frame
        .snapshots().timeRange(timestamp - durationEpsilon, timestamp + durationEpsilon);
        // clang-format on

        // TODO(fabian.reister): remove latest() and add atTime
        // .atTime(timestamp); // transformation

        const armem::client::QueryResult qResult = memoryReader.query(qb.buildQueryInput());

        ARMARX_DEBUG << "Lookup result in reader: " << qResult;

        if (not qResult.success)
        {
            return {.transform =
                {
                    .header = query.header,
                },
                .status       = TransformResult::Status::ErrorFrameNotAvailable,
                .errorMessage = "Error in tf lookup '" + query.header.parentFrame + " -> " +
                query.header.frame + "' : " + qResult.errorMessage};
        }

        const std::vector<std::string> tfChain = buildTransformChain(qResult.memory, query);

        if (tfChain.empty())
        {
            return {.transform    = {.header = query.header},
                    .status       = TransformResult::Status::ErrorFrameNotAvailable,
                    .errorMessage = "Cannot create tf lookup chain '" +
                                    query.header.parentFrame + " -> " + query.header.frame +
                                    "'"};
        }

        const std::vector<Eigen::Affine3f> transforms = obtainTransforms(
                    qResult.memory, tfChain, query.header.agent, timestamp.toMicroSeconds());

        if (transforms.empty())
        {
            ARMARX_WARNING << "No transform available.";
            return {.transform    = {.header = query.header},
                    .status       = TransformResult::Status::ErrorFrameNotAvailable,
                    .errorMessage = "Error in TF loookup:  '" +
                                    query.header.parentFrame + " -> " + query.header.frame +
                                    "'. No memory data in time range."};
        }

        const Eigen::Affine3f transform = std::accumulate(transforms.begin(),
                                          transforms.end(),
                                          Eigen::Affine3f::Identity(),
                                          std::multiplies<>());

        ARMARX_DEBUG << "Found valid transform";

        return {.transform = {.header = query.header, .transform = transform},
                .status    = TransformResult::Status::Success};
    }

} // namespace armarx::armem
