#include "MappingDataWriter.h"

#include "RobotAPI/libraries/armem_robot_mapping/aron_conversions.h"
#include <RobotAPI/libraries/armem_robot_mapping/aron/LaserScan.aron.generated.h>


namespace armarx::armem
{

    MappingDataWriter::MappingDataWriter(ManagedIceObject& component)
        : MemoryConnector(component) {}

    MappingDataWriter::~MappingDataWriter() = default;

    void MappingDataWriter::registerPropertyDefinitions(
        armarx::PropertyDefinitionsPtr& def)
    {
        ARMARX_DEBUG << "TransformWriter: registerPropertyDefinitions";

        MemoryConnector::registerPropertyDefinitions(def);

        const std::string prefix = getPropertyPrefix();

        def->optional(properties.mappingMemoryName, prefix + "MappingMemoryName",
                      "Name of the mapping memory core segment to use.");

        def->optional(properties.memoryName, prefix + "MemoryName");
    }

    void MappingDataWriter::connect()
    {
        // Wait for the memory to become available and add it as dependency.
        ARMARX_IMPORTANT << "MappingDataWriter: Waiting for memory '" << properties.memoryName
                         << "' ...";
        auto result = useMemory(properties.memoryName);
        if (not result.success)
        {
            ARMARX_ERROR << result.errorMessage;
            return;
        }

        ARMARX_IMPORTANT << "TransformWriter: Connected to memory '" << properties.memoryName;

        memoryWriter.setWritingMemory(result.proxy);
    }


    bool MappingDataWriter::storeSensorData(const LaserScan& laserScan,
                                            const std::string& frame,
                                            const std::string& agentName,
                                            const std::int64_t& timestamp)
    {
        std::lock_guard g{memoryWriterMutex};

        const auto result =
            memoryWriter.addSegment(properties.mappingMemoryName, agentName);

        if (not result.success)
        {
            ARMARX_ERROR << result.errorMessage;

            // TODO(fabian.reister): message
            return false;
        }

        const auto iceTimestamp = Time::microSeconds(timestamp);

        const auto providerId = armem::MemoryID(result.segmentID);
        const auto entityID =
            providerId.withEntityName(frame).withTimestamp(iceTimestamp);

        armem::EntityUpdate update;
        update.entityID = entityID;
        update.timeCreated = armem::Time::now();

        arondto::LaserScanStamped aronSensorData;
        // currently only sets the header
        toAron(laserScan, iceTimestamp, frame, agentName, aronSensorData);

        auto dict = aronSensorData.toAron();
        dict->addElement("scan", toAron(laserScan));

        update.instancesData = {dict};
        update.timeCreated = iceTimestamp;

        ARMARX_DEBUG << "Committing " << update << " at time " << iceTimestamp;
        armem::EntityUpdateResult updateResult = memoryWriter.commit(update);

        ARMARX_DEBUG << updateResult;

        if (not updateResult.success)
        {
            ARMARX_ERROR << updateResult.errorMessage;
        }

        return updateResult.success;
    }

} // namespace armarx::armem