/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Header
#include "List.h"

// ArmarX
#include <ArmarXCore/core/exceptions/Exception.h>
#include <RobotAPI/libraries/aron/core/navigator/data/NavigatorFactory.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/List.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/Tuple.h>
#include <RobotAPI/libraries/aron/core/navigator/type/container/Pair.h>

namespace armarx::aron::datanavigator
{
    // constructors
    ListNavigator::ListNavigator(const Path& path) :
        aron::Navigator<data::Descriptor, data::AronData>::Navigator(data::Descriptor::eList, path),
        Navigator(data::Descriptor::eList, path),
        aron(new data::AronList())
    {
    }

    ListNavigator::ListNavigator(const data::AronListPtr& l, const Path& path) :
        aron::Navigator<data::Descriptor, data::AronData>::Navigator(data::Descriptor::eList, path),
        Navigator(data::Descriptor::eList, path),
        aron(l)
    {
        CheckAronPtrForNull("ListNavigator", "ListNavigator", getPath(), aron);

        unsigned int i = 0;
        for (const auto& dataPtr : l->elements)
        {
            childrenNavigators.push_back(FromAronData(dataPtr, Path(path, std::to_string(i++))));
        }
    }

    ListNavigator::ListNavigator(const data::AronDataList& d, const Path& path) :
        ListNavigator(data::AronListPtr(new data::AronList(d)), path)
    {
    }

    ListNavigator::ListNavigator(const std::vector<NavigatorPtr>& n, const Path& path) :
        ListNavigator(path)
    {
        for (const auto& dataPtr : n)
        {
            addElement(dataPtr);
        }
    }

    // operators
    bool ListNavigator::operator==(const ListNavigator& other) const
    {
        unsigned int i = 0;
        for (const auto& nav : childrenNavigators)
        {
            if (not(other.hasElement(i)))
            {
                return false;
            }
            if (not(nav->equalsDataNavigator(other.getElement(i))))
            {
                return false;
            }
            i++;
        }
        return true;
    }

    bool ListNavigator::equalsDataNavigator(const NavigatorPtr& other) const
    {
        if (other == nullptr)
        {
            return false;
        }
        ListNavigatorPtr casted = DynamicCast(other);
        return equalsDeep(casted);
    }

    bool ListNavigator::equalsDeep(const ListNavigatorPtr& other) const
    {
        if (other == nullptr)
        {
            return false;
        }
        return *this == *other;
    }

    // static methods
    ListNavigatorPtr ListNavigator::DynamicCast(const NavigatorPtr& n)
    {
        ListNavigatorPtr casted = std::dynamic_pointer_cast<ListNavigator>(n);
        return casted;
    }

    ListNavigator ListNavigator::DynamicCast(Navigator& n)
    {
        return dynamic_cast<ListNavigator&>(n);
    }

    ListNavigatorPtr ListNavigator::DynamicCastAndCheck(const NavigatorPtr& n)
    {
        CheckDataNavigatorPtrForNull("ListNavigator", "DynamicCastAndCheck[Before]", n);
        ListNavigatorPtr casted = ListNavigator::DynamicCast(n);
        CheckDataNavigatorPtrForNull("ListNavigator", "DynamicCastAndCheck[After]", n->getPath(), casted);
        return casted;
    }

    ListNavigatorPtr ListNavigator::FromAronListPtr(const data::AronListPtr& aron)
    {
        return std::make_shared<ListNavigator>(aron);
    }

    data::AronListPtr ListNavigator::ToAronListPtr(const ListNavigatorPtr& navigator)
    {
        return navigator ? navigator->toAronListPtr() : nullptr;
    }

    // public member functions
    void ListNavigator::addElement(const NavigatorPtr& n)
    {
        childrenNavigators.push_back(n);
        aron->elements.push_back(n->getResult());
    }

    bool ListNavigator::hasElement(unsigned int i) const
    {
        return i < childrenNavigators.size();
    }

    NavigatorPtr ListNavigator::getElement(unsigned int i) const
    {
        if (i >= childrenNavigators.size())
        {
            throw error::AronException("ListNavigator", "getElement", "The index i = " + std::to_string(i) + " is out of bounds (size = " + std::to_string(childrenNavigators.size()) + ")", getPath());
        }
        return childrenNavigators[i];
    }

    std::vector<NavigatorPtr> ListNavigator::getElements() const
    {
        return childrenNavigators;
    }

    void ListNavigator::clear()
    {
        childrenNavigators.clear();
        aron->elements.clear();
    }

    data::AronListPtr ListNavigator::toAronListPtr() const
    {
        return aron;
    }

    // virtual implementations
    data::AronDataPtr ListNavigator::getResult() const
    {
        return toAronListPtr();
    }

    std::string ListNavigator::getName() const
    {
        return "AronList";
    }

    typenavigator::NavigatorPtr ListNavigator::recalculateType() const
    {
        typenavigator::ListNavigatorPtr typenav = typenavigator::ListNavigatorPtr(new typenavigator::ListNavigator(getPath()));
        unsigned int i = 0;
        for (const auto& nav : childrenNavigators)
        {
            if (typenav->getAcceptedType() == nullptr)
            {
                typenav->setAcceptedType(nav->recalculateType());
                continue;
            }

            if (!nav->fullfillsType(typenav->getAcceptedType()))
            {
                throw error::AronException("ListNavigator", "recalculateType", "Cannot recalculate the aronType. Inconsistency found for index: " + std::to_string(i), getPath());
            }
            ++i;
        }
        return typenav->getAcceptedType();
    }

    bool ListNavigator::fullfillsType(const typenavigator::NavigatorPtr& type) const
    {
        if (!Resolver::Correspond(type->getDescriptor(), getDescriptor()))
        {
            return false;
        }

        type::Descriptor typeDesc = type->getDescriptor();
        switch (typeDesc)
        {
            case type::Descriptor::eList:
            {
                typenavigator::ListNavigatorPtr listTypeNav = typenavigator::ListNavigator::DynamicCast(type);
                for (const auto& nav : childrenNavigators)
                {
                    if (!nav->fullfillsType(listTypeNav->getAcceptedType()))
                    {
                        return false;
                    }
                }
                return true;
            }
            case type::Descriptor::eTuple:
            {
                typenavigator::TupleNavigatorPtr tupleTypeNav = typenavigator::TupleNavigator::DynamicCast(type);
                unsigned int i = 0;
                for (const auto& nav : childrenNavigators)
                {
                    if (!nav->fullfillsType(tupleTypeNav->getAcceptedType(i)))
                    {
                        return false;
                    }
                }
                return true;
            }
            case type::Descriptor::ePair:
            {
                typenavigator::PairNavigatorPtr pairTypeNav = typenavigator::PairNavigator::DynamicCast(type);
                if (childrenSize() != 2)
                {
                    return false;
                }
                return childrenNavigators[0]->fullfillsType(pairTypeNav->getFirstAcceptedType()) && childrenNavigators[1]->fullfillsType(pairTypeNav->getSecondAcceptedType());
            }
            default:
                return false;
        }
    }

    std::vector<NavigatorPtr> ListNavigator::getChildren() const
    {
        return childrenNavigators;
    }

    size_t ListNavigator::childrenSize() const
    {
        return childrenNavigators.size();
    }

    NavigatorPtr ListNavigator::navigateAbsolute(const Path& path) const
    {
        if (!path.hasElement())
        {
            throw error::AronException("ListNavigator", "navigate", "Could not navigate without a valid path. The path was empty.");
        }
        unsigned int i = std::stoi(path.getFirstElement());
        if (!hasElement(i))
        {
            throw error::IndexNotValidException("ListNavigator", "navigate", "Could not find an element of a path.", i, childrenSize());
        }

        if (path.size() == 1)
        {
            return childrenNavigators.at(i);
        }
        else
        {
            Path next = path.withDetachedFirstElement();
            return childrenNavigators.at(i)->navigateAbsolute(next);
        }
    }

    NavigatorPtr ListNavigator::navigateRelative(const Path& path) const
    {
        Path absoluteFromHere = path.getWithoutPrefix(getPath());
        return navigateAbsolute(absoluteFromHere);
    }
}
